__author__ = 'Difu Hu <pyats-support@cisco.com;pyats-support-ext@cisco.com>'

import unittest

from unicon import Connection
from unicon.mock.mock_device import MockDevice
from unicon.plugins.sros import service_implementation


class TestSrosPlugin(unittest.TestCase):

    def setUp(self):
        self.md = MockDevice(device_os='sros', state='execute')
        self.joined = lambda string: '\n'.join(string.splitlines())
        self.con = Connection(
            os='sros',
            hostname='COTKON04XR2',
            start=['mock_device_cli --os sros --state connect_ssh'],
            credentials={'default': {'username': 'grpc', 'password': 'nokia'}}
        )
        self.con.connect()

    def test_connect(self):
        self.assertIn('COTKON04XR2#', self.con.spawn.match.match_output)

    def test_mdcli_execute(self):
        cmd = 'show router interface coreloop'
        output = self.con.mdcli_execute(cmd)
        expect = self.md.mock_data['mdcli_execute']['commands'][cmd]
        self.assertEqual(self.joined(output), self.joined(expect))

    def test_mdcli_configure(self):
        cmd = 'router interface coreloop ipv4 primary address 1.1.1.1 prefix-length 32'
        output = self.con.mdcli_configure(cmd, mode='global')
        expect = self.md.mock_data['mdcli_configure_global']['commands'][cmd]
        # self.assertIn(self.joined(expect), self.joined(output))

    def test_mdcli_configure_commit_fail(self):
        cmd = 'router interface coreloop ipv4 primary address 2.2.2.2 prefix-length 32'
        output = self.con.mdcli_configure(cmd)
        expect = self.md.mock_data['mdcli_configure_private']['commands'][cmd]
        commit = self.md.mock_data['mdcli_configure_private']['commands']['commit']
        self.assertIn(self.joined(expect), self.joined(output))
        self.assertIn(self.joined(commit), self.joined(output))

    def test_classiccli_execute(self):
        cmd = 'show router interface coreloop'
        output = self.con.classiccli_execute(cmd)
        expect = self.md.mock_data['classiccli_execute']['commands'][cmd]
        self.assertEqual(self.joined(output), self.joined(expect))

    def test_classiccli_configure(self):
        cmd = 'configure router interface coreloop address 111.1.1.1 255.255.255.255'
        output = self.con.classiccli_configure(cmd)
        expect = self.md.mock_data['classiccli_execute']['commands'][cmd]['response']
        self.assertIn(self.joined(expect), self.joined(output))

    def test_execute_and_cli_engine(self):
        self.con.switch_cli_engine('classiccli')
        engine = self.con.get_cli_engine()
        self.assertEqual(engine, 'classiccli')
        cmd = 'show router interface coreloop'
        output = self.con.execute(cmd)
        expect = self.md.mock_data['classiccli_execute']['commands'][cmd]
        self.assertEqual(self.joined(output), self.joined(expect))

        self.con.switch_cli_engine('mdcli')
        engine = self.con.get_cli_engine()
        self.assertEqual(engine, 'mdcli')
        cmd = 'show router interface coreloop'
        output = self.con.execute(cmd)
        expect = self.md.mock_data['mdcli_execute']['commands'][cmd]
        self.assertEqual(self.joined(output), self.joined(expect))

    def test_configure_and_cli_engine(self):
        self.con.switch_cli_engine('mdcli')
        engine = self.con.get_cli_engine()
        self.assertEqual(engine, 'mdcli')
        self.con.mdcli_configure.mode = 'global'
        cmd = 'router interface coreloop ipv4 primary address 1.1.1.1 prefix-length 32'
        output = self.con.configure(cmd)
        expect = self.md.mock_data['mdcli_configure_global']['commands'][cmd]
        # self.assertIn(self.joined(expect), self.joined(output))

        self.con.switch_cli_engine('classiccli')
        engine = self.con.get_cli_engine()
        self.assertEqual(engine, 'classiccli')
        cmd = 'configure router interface coreloop address 111.1.1.1 255.255.255.255'
        output = self.con.configure(cmd)
        expect = self.md.mock_data['classiccli_execute']['commands'][cmd]['response']
        self.assertIn(self.joined(expect), self.joined(output))


if __name__ == '__main__':
    unittest.main()
